@@ -1,5 +1,5 @@
 /*  XMMS2 - X Music Multiplexer System
- *  Copyright (C) 2003-2011 XMMS2 Team
+ *  Copyright (C) 2003-2023 XMMS2 Team
  *
  *  PLUGINS ARE NOT CONSIDERED TO BE DERIVED WORK !!!
  *
@@ -26,25 +26,28 @@
 #include <locale.h>
 #include <glib.h>
 
-#include "xmms_configuration.h"
-#include "xmmsc/xmmsc_util.h"
-#include "xmmspriv/xmms_plugin.h"
-#include "xmmspriv/xmms_config.h"
-#include "xmmspriv/xmms_playlist.h"
-#include "xmmspriv/xmms_collection.h"
-#include "xmmspriv/xmms_signal.h"
-#include "xmmspriv/xmms_symlink.h"
-#include "xmmspriv/xmms_checkroot.h"
-#include "xmmspriv/xmms_thread_name.h"
-#include "xmmspriv/xmms_medialib.h"
-#include "xmmspriv/xmms_output.h"
-#include "xmmspriv/xmms_ipc.h"
-#include "xmmspriv/xmms_log.h"
-#include "xmmspriv/xmms_sqlite.h"
-#include "xmmspriv/xmms_xform.h"
-#include "xmmspriv/xmms_bindata.h"
-#include "xmmspriv/xmms_utils.h"
-#include "xmmspriv/xmms_visualization.h"
+#include <xmms_configuration.h>
+#include <xmmsc/xmmsc_util.h>
+#include <xmmspriv/xmms_plugin.h>
+#include <xmmspriv/xmms_config.h>
+#include <xmmspriv/xmms_courier.h>
+#include <xmmspriv/xmms_playlist.h>
+#include <xmmspriv/xmms_playlist_updater.h>
+#include <xmmspriv/xmms_collsync.h>
+#include <xmmspriv/xmms_collection.h>
+#include <xmmspriv/xmms_signal.h>
+#include <xmmspriv/xmms_symlink.h>
+#include <xmmspriv/xmms_checkroot.h>
+#include <xmmspriv/xmms_thread_name.h>
+#include <xmmspriv/xmms_medialib.h>
+#include <xmmspriv/xmms_mediainfo.h>
+#include <xmmspriv/xmms_output.h>
+#include <xmmspriv/xmms_ipc.h>
+#include <xmmspriv/xmms_log.h>
+#include <xmmspriv/xmms_xform_object.h>
+#include <xmmspriv/xmms_bindata.h>
+#include <xmmspriv/xmms_utils.h>
+#include <xmmspriv/xmms_visualization.h>
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -58,13 +61,11 @@
  * Forward declarations of the methods in the main object
  */
 static void xmms_main_client_quit (xmms_object_t *object, xmms_error_t *error);
-static GTree *xmms_main_client_stats (xmms_object_t *object, xmms_error_t *error);
-static GList *xmms_main_client_list_plugins (xmms_object_t *main, gint32 type, xmms_error_t *err);
-static void xmms_main_client_hello (xmms_object_t *object, gint protocolver, const gchar *client, xmms_error_t *error);
+static xmmsv_t *xmms_main_client_stats (xmms_object_t *object, xmms_error_t *error);
+static xmmsv_t *xmms_main_client_list_plugins (xmms_object_t *main, gint32 type, xmms_error_t *err);
+static gint64 xmms_main_client_hello (xmms_object_t *object, gint protocolver, const gchar *client, gint64 id, xmms_error_t *error);
 static void install_scripts (const gchar *into_dir);
 static void spawn_script_setup (gpointer data);
-static xmms_xform_object_t *xform_obj;
-static xmms_bindata_t *bindata_obj;
 
 #include "main_ipc.c"
 
@@ -87,8 +88,17 @@ static xmms_bindata_t *bindata_obj;
  */
 struct xmms_main_St {
 	xmms_object_t object;
-	xmms_output_t *output;
-	xmms_visualization_t *vis;
+	xmms_output_t *output_object;
+	xmms_bindata_t *bindata_object;
+	xmms_coll_dag_t *colldag_object;
+	xmms_medialib_t *medialib_object;
+	xmms_playlist_t *playlist_object;
+	xmms_coll_sync_t *collsync_object;
+	xmms_playlist_updater_t *plsupdater_object;
+	xmms_xform_object_t *xform_object;
+	xmms_mediainfo_reader_t *mediainfo_object;
+	xmms_visualization_t *visualization_object;
+	xmms_courier_t *courier_object;
 	time_t starttime;
 };
 
@@ -100,33 +110,133 @@ static GMainLoop *mainloop;
 /** The path of the configfile */
 static gchar *conffile = NULL;
 
+static void
+query_total_size_duration (xmms_main_t *mainobj, xmms_error_t *error,
+                                      int64_t *size, int64_t *duration)
+{
+	xmmsv_t *coll, *universe, *spec, *ret;
+	xmms_medialib_session_t *session;
+
+	/* Fetch the size in bytes and duration in milliseconds for the whole media library */
+	universe = xmmsv_new_coll (XMMS_COLLECTION_TYPE_UNIVERSE);
+
+	coll = xmmsv_new_coll (XMMS_COLLECTION_TYPE_MATCH);
+	xmmsv_coll_attribute_set_string (coll, "field", "status");
+	xmmsv_coll_attribute_set_string (coll, "value", "1");
+	xmmsv_coll_add_operand (coll, universe);
+	xmmsv_unref (universe);
+
+	spec = xmmsv_build_dict (
+		XMMSV_DICT_ENTRY_STR ("type", "metadata"),
+		XMMSV_DICT_ENTRY ("fields", xmmsv_build_list (
+			XMMSV_LIST_ENTRY_STR ("duration"),
+			XMMSV_LIST_ENTRY_STR ("size"),
+			XMMSV_LIST_END)),
+		XMMSV_DICT_ENTRY ("get", xmmsv_build_list (
+			XMMSV_LIST_ENTRY_STR ("field"),
+			XMMSV_LIST_ENTRY_STR ("value"),
+			XMMSV_LIST_END)),
+		XMMSV_DICT_ENTRY_STR ("aggregate", "sum"),
+		XMMSV_DICT_END);
+
+	do {
+		session = xmms_medialib_session_begin_ro (mainobj->medialib_object);
+		ret = xmms_medialib_query (session, coll, spec, error);
+	} while (!xmms_medialib_session_commit (session));
+
+	xmmsv_dict_entry_get_int64 (ret, "size", size);
+	xmmsv_dict_entry_get_int64 (ret, "duration", duration);
+
+	xmmsv_unref (spec);
+	xmmsv_unref (ret);
+}
+
+static void
+query_total_playtime (xmms_main_t *mainobj, xmms_error_t *error,
+                                 int64_t *playtime)
+{
+	xmmsv_t *coll, *universe, *spec, *ret, *value;
+	xmms_medialib_session_t *session;
+	xmmsv_dict_iter_t *iter;
+	const gchar *key;
+
+	/* Fetch the sum of duration clustered by timesplayed, for timesplayed > 0 */
+	universe = xmmsv_new_coll (XMMS_COLLECTION_TYPE_UNIVERSE);
+
+	coll = xmmsv_new_coll (XMMS_COLLECTION_TYPE_GREATER);
+	xmmsv_coll_attribute_set_string (coll, "field", "timesplayed");
+	xmmsv_coll_attribute_set_string (coll, "value", "0");
+	xmmsv_coll_add_operand (coll, universe);
+	xmmsv_unref (universe);
+
+	spec = xmmsv_build_dict (
+		XMMSV_DICT_ENTRY_STR ("type", "cluster-dict"),
+		XMMSV_DICT_ENTRY_STR ("cluster-by", "value"),
+		XMMSV_DICT_ENTRY_STR ("cluster-field", "timesplayed"),
+		XMMSV_DICT_ENTRY ("data", xmmsv_build_dict (
+			XMMSV_DICT_ENTRY_STR ("type", "metadata"),
+			XMMSV_DICT_ENTRY ("fields", xmmsv_build_list (
+				XMMSV_LIST_ENTRY_STR ("duration"),
+				XMMSV_LIST_END)),
+			XMMSV_DICT_ENTRY ("get", xmmsv_build_list (
+				XMMSV_LIST_ENTRY_STR ("value"),
+				XMMSV_LIST_END)),
+			XMMSV_DICT_ENTRY_STR ("aggregate", "sum"),
+			XMMSV_DICT_END)),
+		XMMSV_DICT_END);
+
+	do {
+		session = xmms_medialib_session_begin_ro (mainobj->medialib_object);
+		ret = xmms_medialib_query (session, coll, spec, error);
+	} while (!xmms_medialib_session_commit (session));
+
+	xmmsv_get_dict_iter (ret, &iter);
+	while (xmmsv_dict_iter_pair (iter, &key, &value)) {
+		int64_t sum, timesplayed;
+		gchar *endptr = NULL;
+
+		if (xmmsv_get_int64 (value, &sum)) {
+			timesplayed = strtol (key, &endptr, 10);
+			if (*endptr == '\0')
+				*playtime += timesplayed * sum;
+		}
+
+		xmmsv_dict_iter_next (iter);
+	}
+
+	xmmsv_unref (spec);
+	xmmsv_unref (ret);
+}
+
 /**
  * This returns the main stats for the server
  */
-static GTree *
+static xmmsv_t *
 xmms_main_client_stats (xmms_object_t *object, xmms_error_t *error)
 {
-	GTree *ret;
-	gint starttime;
+	xmms_main_t *mainobj = (xmms_main_t *) object;
+	gint uptime = time (NULL) - mainobj->starttime;
+	int64_t size, duration, playtime;
 
-	ret = g_tree_new_full ((GCompareDataFunc) strcmp, NULL,
-	                       NULL, (GDestroyNotify) xmmsv_unref);
+	size = duration = playtime = 0;
 
-	starttime = ((xmms_main_t*)object)->starttime;
+	query_total_playtime (mainobj, error, &playtime);
+	query_total_size_duration (mainobj, error, &size, &duration);
 
-	g_tree_insert (ret, (gpointer) "version",
-	               xmmsv_new_string (XMMS_VERSION));
-	g_tree_insert (ret, (gpointer) "uptime",
-	               xmmsv_new_int (time (NULL) - starttime));
-
-	return ret;
+	return xmmsv_build_dict (XMMSV_DICT_ENTRY_STR ("version", XMMS_VERSION),
+	                         XMMSV_DICT_ENTRY_INT ("uptime", uptime),
+	                         XMMSV_DICT_ENTRY_INT ("size", size),
+	                         XMMSV_DICT_ENTRY_INT ("duration", duration),
+	                         XMMSV_DICT_ENTRY_INT ("playtime", playtime),
+	                         XMMSV_DICT_END);
 }
 
 static gboolean
 xmms_main_client_list_foreach (xmms_plugin_t *plugin, gpointer data)
 {
-	xmmsv_t *dict;
-	GList **list = data;
+	xmmsv_t *list, *dict;
+
+	list = (xmmsv_t *) data;
 
 	dict = xmmsv_build_dict (
 	        XMMSV_DICT_ENTRY_STR ("name", xmms_plugin_name_get (plugin)),
@@ -136,16 +246,17 @@ xmms_main_client_list_foreach (xmms_plugin_t *plugin, gpointer data)
 	        XMMSV_DICT_ENTRY_INT ("type", xmms_plugin_type_get (plugin)),
 	        XMMSV_DICT_END);
 
-	*list = g_list_prepend (*list, dict);
+	xmmsv_list_append (list, dict);
+	xmmsv_unref (dict);
 
 	return TRUE;
 }
 
-static GList *
+static xmmsv_t *
 xmms_main_client_list_plugins (xmms_object_t *main, gint32 type, xmms_error_t *err)
 {
-	GList *list = NULL;
-	xmms_plugin_foreach (type, xmms_main_client_list_foreach, &list);
+	xmmsv_t *list = xmmsv_new_list ();
+	xmms_plugin_foreach (type, xmms_main_client_list_foreach, list);
 	return list;
 }
 
@@ -243,7 +354,7 @@ change_output (xmms_object_t *object, xmmsv_t *_data, gpointer userdata)
 	xmms_main_t *mainobj = (xmms_main_t*)userdata;
 	const gchar *outname;
 
-	if (!mainobj->output)
+	if (!mainobj->output_object)
 		return;
 
 	outname = xmms_config_property_get_string ((xmms_config_property_t *) object);
@@ -254,7 +365,7 @@ change_output (xmms_object_t *object, xmmsv_t *_data, gpointer userdata)
 	if (!plugin) {
 		xmms_log_error ("Baaaaad output plugin, try to change the output.plugin config variable to something useful");
 	} else {
-		if (!xmms_output_plugin_switch (mainobj->output, plugin)) {
+		if (!xmms_output_plugin_switch (mainobj->output_object, plugin)) {
 			xmms_log_error ("Baaaaad output plugin, try to change the output.plugin config variable to something useful");
 		}
 	}
@@ -268,25 +379,22 @@ static void
 xmms_main_destroy (xmms_object_t *object)
 {
 	xmms_main_t *mainobj = (xmms_main_t *) object;
-	xmms_object_cmd_arg_t arg;
 	xmms_config_property_t *cv;
 
 	cv = xmms_config_lookup ("core.shutdownpath");
 	do_scriptdir (xmms_config_property_get_string (cv), "stop");
 
-	/* stop output */
-	xmms_object_cmd_arg_init (&arg);
-	arg.args = xmmsv_new_list ();
-	xmms_object_cmd_call (XMMS_OBJECT (mainobj->output),
-	                      XMMS_IPC_CMD_STOP, &arg);
-	xmmsv_unref (arg.args);
-
-	g_usleep (G_USEC_PER_SEC); /* wait for the output thread to end */
-
-	xmms_object_unref (mainobj->vis);
-	xmms_object_unref (mainobj->output);
-
-	xmms_object_unref (xform_obj);
+	xmms_object_unref (mainobj->xform_object);
+	xmms_object_unref (mainobj->visualization_object);
+	xmms_object_unref (mainobj->output_object);
+	xmms_object_unref (mainobj->bindata_object);
+	xmms_object_unref (mainobj->playlist_object);
+	xmms_object_unref (mainobj->colldag_object);
+	xmms_object_unref (mainobj->medialib_object);
+	xmms_object_unref (mainobj->mediainfo_object);
+	xmms_object_unref (mainobj->plsupdater_object);
+	xmms_object_unref (mainobj->collsync_object);
+	xmms_object_unref (mainobj->courier_object);
 
 	xmms_config_save ();
 
@@ -304,23 +412,27 @@ xmms_main_destroy (xmms_object_t *object)
 /**
  * @internal Function to respond to the 'hello' sent from clients on connect
  */
-static void
-xmms_main_client_hello (xmms_object_t *object, gint protocolver, const gchar *client, xmms_error_t *error)
+static gint64
+xmms_main_client_hello (xmms_object_t *object, gint protocolver, const gchar *client, gint64 id, xmms_error_t *error)
 {
 	if (protocolver != XMMS_IPC_PROTOCOL_VERSION) {
 		xmms_log_info ("Client '%s' with bad protocol version (%d, not %d) connected", client, protocolver, XMMS_IPC_PROTOCOL_VERSION);
 		xmms_error_set (error, XMMS_ERROR_INVAL, "Bad protocol version");
-		return;
+	} else {
+		XMMS_DBG ("Client '%s' connected", client);
 	}
-	XMMS_DBG ("Client '%s' connected", client);
+
+	return id;
 }
 
 static gboolean
 kill_server (gpointer object) {
-	xmms_object_emit_f (XMMS_OBJECT (object),
-	                    XMMS_IPC_SIGNAL_QUIT,
-	                    XMMSV_TYPE_INT32,
-	                    time (NULL)-((xmms_main_t*)object)->starttime);
+	xmms_main_t *mainobj = (xmms_main_t *) object;
+	gint uptime = time (NULL) - mainobj->starttime;
+
+	xmms_object_emit (XMMS_OBJECT (object),
+	                  XMMS_IPC_SIGNAL_MAIN_QUIT,
+	                  xmmsv_new_int (uptime));
 
 	xmms_object_unref (object);
 
@@ -387,7 +499,7 @@ static void
 print_version (void)
 {
 	printf ("XMMS2 version " XMMS_VERSION "\n");
-	printf ("Copyright (C) 2003-2011 XMMS2 Team\n");
+	printf ("Copyright (C) 2003-2023 XMMS2 Team\n");
 	printf ("This is free software; see the source for copying conditions.\n");
 	printf ("There is NO warranty; not even for MERCHANTABILITY or FITNESS FOR A\n");
 	printf ("PARTICULAR PURPOSE.\n");
@@ -398,7 +510,6 @@ print_version (void)
 	        glib_major_version,
 	        glib_minor_version,
 	        glib_micro_version);
-	xmms_sqlite_print_version ();
 
 	exit (EXIT_SUCCESS);
 }
@@ -413,7 +524,6 @@ main (int argc, char **argv)
 	xmms_config_property_t *cv;
 	xmms_main_t *mainobj;
 	int loglevel = 1;
-	xmms_playlist_t *playlist;
 	gchar default_path[XMMS_PATH_MAX + 16], *tmp;
 	gboolean verbose = FALSE;
 	gboolean quiet = FALSE;
@@ -422,7 +532,7 @@ main (int argc, char **argv)
 	gboolean showhelp = FALSE;
 	const gchar *outname = NULL;
 	const gchar *ipcpath = NULL;
-	gchar *ppath = NULL;
+	gchar *uuid, *ppath = NULL;
 	int status_fd = -1;
 	GOptionContext *context = NULL;
 	GError *error = NULL;
@@ -503,8 +613,6 @@ main (int argc, char **argv)
 		print_version ();
 	}
 
-	g_thread_init (NULL);
-
 	g_random_set_seed (time (NULL));
 
 	xmms_log_init (loglevel);
@@ -539,15 +647,28 @@ main (int argc, char **argv)
 	}
 
 	if (!xmms_plugin_init (ppath)) {
-		return 1;
+		exit (EXIT_FAILURE);
 	}
 
-	playlist = xmms_playlist_init ();
-	xform_obj = xmms_xform_object_init ();
-	bindata_obj = xmms_bindata_init ();
-
 	mainobj = xmms_object_new (xmms_main_t, xmms_main_destroy);
 
+	mainobj->medialib_object = xmms_medialib_init ();
+	mainobj->colldag_object = xmms_collection_init (mainobj->medialib_object);
+	mainobj->mediainfo_object = xmms_mediainfo_reader_start (mainobj->medialib_object);
+	mainobj->playlist_object = xmms_playlist_init (mainobj->medialib_object,
+	                                               mainobj->colldag_object);
+
+	uuid = xmms_medialib_uuid (mainobj->medialib_object);
+	mainobj->collsync_object = xmms_coll_sync_init (uuid,
+	                                                mainobj->colldag_object,
+	                                                mainobj->playlist_object);
+	g_free (uuid);
+	mainobj->plsupdater_object = xmms_playlist_updater_init (mainobj->playlist_object);
+
+	mainobj->xform_object = xmms_xform_object_init ();
+	mainobj->bindata_object = xmms_bindata_init ();
+	mainobj->courier_object = xmms_courier_init();
+
 	/* find output plugin. */
 	cv = xmms_config_property_register ("output.plugin",
 	                                    XMMS_OUTPUT_DEFAULT,
@@ -559,18 +680,20 @@ main (int argc, char **argv)
 
 	outname = xmms_config_property_get_string (cv);
 	xmms_log_info ("Using output plugin: %s", outname);
-	o_plugin = (xmms_output_plugin_t *)xmms_plugin_find (XMMS_PLUGIN_TYPE_OUTPUT, outname);
+	o_plugin = (xmms_output_plugin_t *) xmms_plugin_find (XMMS_PLUGIN_TYPE_OUTPUT, outname);
 	if (!o_plugin) {
 		xmms_log_error ("Baaaaad output plugin, try to change the"
 		                "output.plugin config variable to something useful");
 	}
 
-	mainobj->output = xmms_output_new (o_plugin, playlist);
-	if (!mainobj->output) {
+	mainobj->output_object = xmms_output_new (o_plugin,
+	                                          mainobj->playlist_object,
+	                                          mainobj->medialib_object);
+	if (!mainobj->output_object) {
 		xmms_log_fatal ("Failed to create output object!");
 	}
 
-	mainobj->vis = xmms_visualization_new (mainobj->output);
+	mainobj->visualization_object = xmms_visualization_new (mainobj->output_object);
 
 	if (status_fd != -1) {
 		write (status_fd, "+", 1);
@@ -613,7 +736,7 @@ main (int argc, char **argv)
 
 	g_main_loop_run (mainloop);
 
-	return 0;
+	return EXIT_SUCCESS;
 }
 
 /** @} */
